@@ -28,6 +28,10 @@
 #define BUFSIZ 1024
 #endif
 
+#ifdef VMS
+#define getenv _getenv
+#endif
+
 extern char *getenv();
 
 #include <X11/Intrinsic.h>
@@ -40,7 +44,7 @@ extern char *getenv();
 #include <X11/Xaw/Command.h>
 
 #include "Ghostview.h"
-#include "ghostview.h"
+#include "gv.h"
 #include "ps.h"
 
 /* Start application folding up by Destroying the top level widget. */
@@ -145,6 +149,24 @@ popup_dialog(w, client_data, call_data)
     }
 }
 
+/* Explicitly reopen the file. */
+void
+reopen_file(w, client_data, call_data)
+    Widget w;
+    XtPointer client_data, call_data;
+{
+    struct stat sbuf;
+    int number = current_page;
+
+    fclose(psfile);
+    psfile = fopen(filename, "r");
+    mtime = sbuf.st_mtime;
+    if (oldfilename) XtFree(oldfilename);
+    oldfilename = XtNewString(filename);
+    new_file(number);
+    show_page(number);
+}
+
 /* Get the selection, if no selection, get the insertion point. */
 /* If the new_page is different from the current page show it.  */
 /* If not at the first page, show the previous page. */
@@ -488,7 +510,8 @@ track_and_zoom(w, client_data, call_data)
     zoompage = XtCreateManagedWidget("page", ghostviewWidgetClass,
 				     zoomform, args, num_args);
     num_ghosts++;
-    XtAddCallback(zoompage, XtNcallback, track_and_zoom, (XtPointer)zoom);
+    XtAddCallback(zoompage, XtNcallback, track_and_zoom, (XtPointer)0);
+    XtAddCallback(zoompage, XtNmessageCallback, message, (XtPointer)zoompage);
     XtAddCallback(zoompage, XtNdestroyCallback, destroy_ghost,
 		  (XtPointer)zoompage);
 
@@ -504,7 +527,7 @@ track_and_zoom(w, client_data, call_data)
     XtSetArg(args[num_args], XtNright, XtChainRight);	num_args++;
     zoomdismiss = XtCreateManagedWidget("dismiss", commandWidgetClass,
 				       zoomform, args, num_args);
-    XtAddCallback(zoomdismiss, XtNcallback, dismiss, (XtPointer)zoom);
+    XtAddCallback(zoomdismiss, XtNcallback, destroy, (XtPointer)zoom);
 
     XtSetArg(args[0], XtNwidth, &width);
     XtGetValues(zoompage, args, ONE);
@@ -532,15 +555,15 @@ track_and_zoom(w, client_data, call_data)
 	zoomfile = fopen(filename, "r");
 	if (zoomfile == NULL) return;
 	GhostviewSendPS(zoompage, zoomfile, doc->beginprolog,
-			doc->endprolog, False);
+			doc->lenprolog, False);
 	GhostviewSendPS(zoompage, zoomfile, doc->beginsetup,
-			doc->endsetup, False);
+			doc->lensetup, False);
 	if (doc->pageorder == DESCEND)
 	    i = (doc->numpages - 1) - current_page;
 	else
 	    i = current_page;
 	GhostviewSendPS(zoompage, zoomfile, doc->pages[i].begin,
-			doc->pages[i].end, True);
+			doc->pages[i].len, True);
     }
 }
 
@@ -552,21 +575,36 @@ message(w, client_data, call_data)
     XtPointer client_data, call_data;
 {
     int i;
+    char *error;
 
     if (!strcmp((char *) call_data, "Failed")) {
-	output(page, NULL, "Error: PostScript interpreter failed.\n\n");
+	if ((Widget)client_data == page) {
+	    error = "Error: PostScript interpreter failed in main window.\n\n";
+	} else {
+	    error = "Error: PostScript interpreter failed in zoom window.\n\n";
+	}
+	output(w, NULL, error);
+    } else if (!strcmp((char *) call_data, "BadAlloc")) {
+	if ((Widget)client_data == page) {
+	    error = 
+	       "Warning: Could not allocate backing pixmap in main window.\n\n";
+	} else {
+	    error = 
+	       "Warning: Could not allocate backing pixmap in zoom window.\n\n";
+	}
+	output(w, NULL, error);
     } else if (!strcmp((char *) call_data, "Refresh")) {
 	if (toc_text) {
-	    GhostviewSendPS(page, psfile, doc->beginprolog,
-			    doc->endprolog, False);
-	    GhostviewSendPS(page, psfile, doc->beginsetup,
-			    doc->endsetup, False);
+	    GhostviewSendPS(w, psfile, doc->beginprolog,
+			    doc->lenprolog, False);
+	    GhostviewSendPS(w, psfile, doc->beginsetup,
+			    doc->lensetup, False);
 	    if (doc->pageorder == DESCEND)
 		i = (doc->numpages - 1) - current_page;
 	    else
 		i = current_page;
-	    GhostviewSendPS(page, psfile, doc->pages[i].begin,
-			    doc->pages[i].end, False);
+	    GhostviewSendPS(w, psfile, doc->pages[i].begin,
+			    doc->pages[i].len, False);
 	}
     }
 }
@@ -663,6 +701,15 @@ dismiss(w, client_data, call_data)
     if ((Widget)client_data == infopopup) info_up = False;
 }
 
+/* Destroy popup window */
+void
+destroy(w, client_data, call_data)
+    Widget w;
+    XtPointer client_data, call_data;
+{
+    XtDestroyWidget((Widget)client_data);
+}
+
 /* destroy callback for Ghostview widgets. */
 /* The disable interpreter call ensures that ghostscript is killed. */
 /* One the count goes to 0, we are sure that all forked processes have */
@@ -675,6 +722,7 @@ destroy_ghost(w, client_data, call_data)
     GhostviewDisableInterpreter((Widget) client_data);
     num_ghosts--;
     if (num_ghosts) return;
+    if (dying) old_Xerror(XtDisplay(w), &bomb);
     XtDestroyApplicationContext(app_con);
     exit(0);
 }
